package cn.kdan.compdf.service.impl;

import cn.hutool.core.util.ObjectUtil;
import cn.hutool.json.JSONUtil;
import cn.kdan.compdf.base.R;
import cn.kdan.compdf.config.JwtConfig;
import cn.kdan.compdf.config.MyRedisConstant;
import cn.kdan.compdf.constant.UserBalanceConstant;
import cn.kdan.compdf.dto.*;
import cn.kdan.compdf.entity.BackgroundUser;
import cn.kdan.compdf.enums.AuthEnum;
import cn.kdan.compdf.enums.RabbitMQEnum;
import cn.kdan.compdf.enums.ResponseEnum;
import cn.kdan.compdf.exception.BusinessException;
import cn.kdan.compdf.mapper.BackgroundUserMapper;
import cn.kdan.compdf.service.BackgroundUserBalanceService;
import cn.kdan.compdf.service.BackgroundUserProjectService;
import cn.kdan.compdf.service.BackgroundUserRemainService;
import cn.kdan.compdf.service.BackgroundUserService;
import cn.kdan.compdf.utils.MD5Util;
import cn.kdan.compdf.utils.TokenUtil;
import cn.kdan.compdf.vo.LoginVO;
import com.baomidou.mybatisplus.core.conditions.query.LambdaQueryWrapper;
import com.baomidou.mybatisplus.core.metadata.IPage;
import com.baomidou.mybatisplus.core.toolkit.Wrappers;
import com.baomidou.mybatisplus.extension.plugins.pagination.Page;
import com.baomidou.mybatisplus.extension.service.impl.ServiceImpl;
import lombok.RequiredArgsConstructor;
import lombok.extern.slf4j.Slf4j;
import org.springframework.amqp.rabbit.core.RabbitTemplate;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.data.redis.core.StringRedisTemplate;
import org.springframework.stereotype.Service;
import org.springframework.web.bind.annotation.GetMapping;

import java.time.LocalDateTime;
import java.util.Date;
import java.util.Objects;
import java.util.concurrent.TimeUnit;

/**
 * @author comPDF-Kit WPH 2022-10-09
 */
@Slf4j
@Service
@RequiredArgsConstructor
public class BackgroundUserServiceImpl extends ServiceImpl<BackgroundUserMapper, BackgroundUser> implements BackgroundUserService {

    @Autowired
    private StringRedisTemplate redisTemplate;
    @Autowired
    private BackgroundUserProjectService backgroundUserProjectService;

    private final BackgroundUserRemainService userRemainService;
    @Autowired
    private BackgroundUserBalanceService userBalanceService;

    private final RabbitTemplate rabbitTemplate;

    @Override
    public IPage<BackgroundUser> page(BackgroundUser query) {
        return this.page(new Page<>(query.getPage(), query.getSize()), Wrappers.query(query));
    }

    @Override
    public R<LoginVO> login(LoginDTO loginDTO) {
        BackgroundUser backgroundUser = this.baseMapper.selectOne(new LambdaQueryWrapper<BackgroundUser>().eq(BackgroundUser::getUsername, loginDTO.getUsername()));
        if (null == backgroundUser) {
            throw new BusinessException("账号不存在");
        }
        //转换成MD5密码
        String md5Password = MD5Util.MD5Encode(loginDTO.getPassword());
        if (!md5Password.equals(backgroundUser.getPassword())) {
            throw new BusinessException("密码错误");
        }
        String token = JwtConfig.generateJwt(LoginUserDTO
                .builder()
                .id(backgroundUser.getId())
                .username(backgroundUser.getUsername())
                .fullName(backgroundUser.getFullName())
                .tenantId(backgroundUser.getTenantId())
                .build());
        log.info("用户登录成功：username：{}，time：{}", loginDTO.getUsername(), new Date());

        // 设置token过期时间
        redisTemplate.opsForValue().set(AuthEnum.USER_TOKEN_REDIS + loginDTO.getUsername(), token, AuthEnum.EXPIRE, TimeUnit.HOURS);

        //更新登录用户的登陆时间
        if (backgroundUser.getFirstLoginTime() == null) {
            // 首次登录创建默认项目
            backgroundUserProjectService.defaultProject(backgroundUser.getId(), backgroundUser.getTenantId());
            backgroundUser.setFirstLoginTime(LocalDateTime.now());
        }
        backgroundUser.setLastLoginTime(new Date());
        baseMapper.updateById(backgroundUser);

        LoginVO loginVO = new LoginVO();
        loginVO.setId(backgroundUser.getId());
        loginVO.setUsername(backgroundUser.getUsername());
        loginVO.setFirstLoginTime(backgroundUser.getFirstLoginTime().toString());
        loginVO.setToken(token);
        return R.ok(loginVO);
    }

    @Override
    public void logout(String token) {
        redisTemplate.delete(AuthEnum.USER_TOKEN_REDIS + token);
    }

    @Override
    public void changePassword(ChangePasswordDTO changePasswordDTO, String token) {
        if (changePasswordDTO.getOldPassword().equals(changePasswordDTO.getNewPassword())) {
            throw new BusinessException(ResponseEnum.OLDPWD_EQUALS_NEWPWD);
        }
        // 获取当前登录用户
        LoginUserDTO requestHeader = TokenUtil.getRequestHeader();
        BackgroundUser backgroundUser = this.baseMapper.selectOne(new LambdaQueryWrapper<BackgroundUser>().eq(BackgroundUser::getId, requestHeader.getId()));
        //转换成MD5密码
        String oldPassword = MD5Util.MD5Encode(changePasswordDTO.getOldPassword());
        if (!oldPassword.equals(backgroundUser.getPassword())) {
            throw new BusinessException("原密码错误");
        }
        //转换成MD5密码
        String newPassword = MD5Util.MD5Encode(changePasswordDTO.getNewPassword());
        backgroundUser.setPassword(newPassword);
        baseMapper.updateById(backgroundUser);
        // 退出登录
        redisTemplate.delete(AuthEnum.USER_TOKEN_REDIS + token);
    }

    @Override
    public Long getUserId(Long tenantId) {
        return this.baseMapper.selectOne(new LambdaQueryWrapper<BackgroundUser>().select(BackgroundUser::getId).eq(BackgroundUser::getTenantId, tenantId)).getId();
    }

    @Override
    public LocalDateTime getUSerFirstLogin(Long tenantId) {
        return this.baseMapper.selectOne(new LambdaQueryWrapper<BackgroundUser>().select(BackgroundUser::getFirstLoginTime)
                .eq(BackgroundUser::getTenantId, tenantId)).getFirstLoginTime();
    }

    @Override
    public void userInit(Long userId, Long tenantId, Integer assets) {
        // 初始化Free账单
        userBalanceService.initUserBalance(userId, tenantId, assets);
        // 初始化Free Remain
        userRemainService.initFreeRemain(userId, tenantId, assets);
        // 消息队列发送到转档同步数据 Free
        AssetRechargeDTO assetRechargeDTO = new AssetRechargeDTO();
        assetRechargeDTO.setTenantId(tenantId);
        assetRechargeDTO.setAssetType(UserBalanceConstant.DESCRIPTION_TYPE_FREE);
        assetRechargeDTO.setAsset(assets);
        String jsonStr = JSONUtil.toJsonStr(assetRechargeDTO);
        rabbitTemplate.convertAndSend(RabbitMQEnum.BACKGROUND_EXCHANGE,
                RabbitMQEnum.BACKGROUND_USER_ASSET_RECHARGE_ROUTING_KEY,
                jsonStr);
        // 首次登录创建默认项目
        backgroundUserProjectService.defaultProject(userId, tenantId);
        BackgroundUser backgroundUser = new BackgroundUser();
        backgroundUser.setFirstLoginTime(LocalDateTime.now());
        backgroundUser.setId(userId);
        this.baseMapper.updateById(backgroundUser);
    }

    @Override
    public LoginUserDTO getUserInfo(LoginUser loginUser) {
        BackgroundUser backgroundUser = this.baseMapper.selectOne(new LambdaQueryWrapper<BackgroundUser>()
                .eq(BackgroundUser::getCompdfkitId, loginUser.getCompdfkitId()));
        if (ObjectUtil.isEmpty(backgroundUser)) {
            Boolean absent = redisTemplate.opsForValue()
                    .setIfAbsent(MyRedisConstant.USER_CREAT+loginUser.toString(),"",10, TimeUnit.MINUTES);
            if (ObjectUtil.isNotEmpty(absent) && absent){
                return this.creatUser(loginUser);
            }else {
                return getUserInfo(loginUser);
            }
        }
        if (!loginUser.getUsername().equals(backgroundUser.getUsername())) {
            // 修改用户名为新的email
            backgroundUser.setUsername(loginUser.getUsername());
            this.baseMapper.updateById(backgroundUser);
            UserUpdDTO userUpdDTO = new UserUpdDTO();
            userUpdDTO.setTenantId(backgroundUser.getTenantId());
            userUpdDTO.setEmail(backgroundUser.getUsername());
            String jsonStr = JSONUtil.toJsonStr(userUpdDTO);
            rabbitTemplate.convertAndSend(RabbitMQEnum.BACKGROUND_EXCHANGE,
                    RabbitMQEnum.BACKGROUND_USER_UPD_ROUTING_KEY,
                    jsonStr);
        }
        LoginUserDTO userDTO = new LoginUserDTO();
        userDTO.setUsername(loginUser.getUsername());
        userDTO.setId(backgroundUser.getId());
        userDTO.setTenantId(backgroundUser.getTenantId());
        userDTO.setFullName(backgroundUser.getFullName());
        userDTO.setUsersId(Math.toIntExact(backgroundUser.getCompdfkitId()));
        return userDTO;
    }

    @Override
    public LoginUserDTO creatUser(LoginUser loginUser) {

        BackgroundUser backgroundUser = new BackgroundUser();
        backgroundUser.setUsername(loginUser.getUsername());
        backgroundUser.setCompdfkitId(loginUser.getCompdfkitId());
        this.baseMapper.insert(backgroundUser);
        // 设置tenantId和主键id保持一致
        backgroundUser.setTenantId(backgroundUser.getId());
        this.baseMapper.updateById(backgroundUser);
        //  添加用户信息(数据同步到SaaS转档服务)
        UserCreatDTO userCreatDTO = new UserCreatDTO();
        userCreatDTO.setTenantId(backgroundUser.getTenantId());
        userCreatDTO.setEmail(backgroundUser.getUsername());
        String jsonStr = JSONUtil.toJsonStr(userCreatDTO);
        rabbitTemplate.convertAndSend(RabbitMQEnum.BACKGROUND_EXCHANGE,
                RabbitMQEnum.BACKGROUND_USER_CREAT_ROUTING_KEY,
                jsonStr);
        // 初始化用户账单个人信息
        this.userInit(backgroundUser.getId(), backgroundUser.getTenantId(), loginUser.getAssets());

        return LoginUserDTO.builder()
                .id(backgroundUser.getId())
                .tenantId(backgroundUser.getTenantId())
                .username(backgroundUser.getUsername())
                .fullName(backgroundUser.getFullName())
                .usersId(Math.toIntExact(backgroundUser.getCompdfkitId()))
                .build();
    }

    @Override
    public Long getCompdfkitId(Long tenantId) {
        BackgroundUser backgroundUser = this.baseMapper.selectOne(new LambdaQueryWrapper<BackgroundUser>().select(BackgroundUser::getCompdfkitId)
                .eq(BackgroundUser::getTenantId, tenantId));
        if (Objects.isNull(backgroundUser)){
            return null;
        }
        return backgroundUser.getCompdfkitId();
    }

    @GetMapping("/verifyUser")
    public R<Void> verifyUser() {
        return R.ok();
    }

}
